def read_file(filename):
    with open(filename, 'r') as file:
        lines = [line.strip().strip('[]').split(',') for line in file.read().splitlines()]
    return [[int(node.strip()) for node in line] for line in lines]

def find_all_violations(sublist, combined_list):
    combined_index = {node: i for i, node in enumerate(combined_list)}
    violations = []
    for i in range(len(sublist) - 1):
        if sublist[i] not in combined_index or sublist[i + 1] not in combined_index:
            violations.append((sublist[i], sublist[i + 1], "Node not found in combined list"))
        elif combined_index[sublist[i]] > combined_index[sublist[i + 1]]:
            violations.append((sublist[i], sublist[i + 1], "Order violation"))
    return violations

def check_order_preservation(correct_lists):
    if len(correct_lists) < 3:
        print("The file does not contain three lists.")
        return
    
    list1 = correct_lists[0]
    list2 = correct_lists[1]
    combined_list = correct_lists[2]

    list1_violations = find_all_violations(list1, combined_list)
    list2_violations = find_all_violations(list2, combined_list)

    if not list1_violations and not list2_violations:
        print("The order of nodes in the first two lists is preserved in the third list.")
    else:
        if list1_violations:
            print("The order of nodes in the first list is not preserved in the third list:")
            for node1, node2, reason in list1_violations:
                print(f"Node {node1} should come before Node {node2}. Reason: {reason}")
        if list2_violations:
            print("The order of nodes in the second list is not preserved in the third list:")
            for node1, node2, reason in list2_violations:
                print(f"Node {node1} should come before Node {node2}. Reason: {reason}")

def main():
    correct_file = 'mytests/solutions/korrekt12.sol'

    correct_lists = read_file(correct_file)

    check_order_preservation(correct_lists)

if __name__ == "__main__":
    main()
