import os
import logging
import time
from datetime import datetime, timedelta
from pulp import *
# Erstellen eines Graphen zur Bestimmung der Knotenreihenfolge
from collections import defaultdict, deque


# Erstelle einen FileHandler und einen StreamHandler
# Zykel einzeln, ohne bestimmte reihenfolge, jedoch mit existenzprüfung für die constraints
file_handler = logging.FileHandler('cycle_einzeln_constraintüberprüfung.log')
console_handler = logging.StreamHandler()
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s', handlers=[file_handler, console_handler])

def count_crossings_via_variables(c_vars):
    crossings = 0
    for c_var in c_vars.values():
        if c_var.varValue == 1:
            crossings += 1
    return(crossings)

def detect_cycle(graph, node, visited, rec_stack, path):
    visited[node] = True
    rec_stack[node] = True
    path.append(node)

    for neighbor in graph[node]:
        if not visited[neighbor]:
            if detect_cycle(graph, neighbor, visited, rec_stack, path):
                return True
        elif rec_stack[neighbor]:
            path.append(neighbor)
            return True

    rec_stack[node] = False
    path.pop()
    return False

def has_cycle(graph, nodes):
    visited = {node: False for node in nodes}
    rec_stack = {node: False for node in nodes}
    path = []
    
    for node in nodes:
        if not visited[node]:
            if detect_cycle(graph, node, visited, rec_stack, path):
                cycle_start = path.index(path[-1])
                return True, path[cycle_start:]
    
    return False, []

"""def add_cycle_constraints(prob, y, cycle,c):
    # Entferne Duplikate aus dem Zyklus
    cycle = list(dict.fromkeys(cycle))
    counter = 0
    for i in cycle:
        for j in cycle:
            for k in cycle:
                if i != j and j != k and i != k:
                    if (i, j) in y and (j, k) in y:
                        prob += 0 <= y[(i, j)] + y[(j, k)] - y[(i, k)] <= 1
                        counter += 1
    a = count_crossings_via_variables(c)
    logging.info(f"Crossings: {a}")
    logging.info(f"Es wurden {counter} Transitivitätsconstraints aufgestellt.")"""
def add_cycle_constraints(prob, y, cycle , added_constraints):
    cycle = list(dict.fromkeys(cycle))
    for i in range(len(cycle)):
        for j in range(i + 1, len(cycle)):
            if (cycle[i], cycle[j]) not in added_constraints:
                prob += y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[i])] == 1
                added_constraints.add((cycle[i], cycle[j]))
                added_constraints.add((cycle[j], cycle[i]))
                
            for k in range(j + 1, len(cycle)):
                if (cycle[i], cycle[j]) in y and (cycle[j], cycle[k]) in y and (cycle[i], cycle[k]) in y:
                    constraint_1 = (cycle[i], cycle[j], cycle[j], cycle[k], cycle[i], cycle[k])
                    constraint_2 = (cycle[i], cycle[k], cycle[j], cycle[j], cycle[i], cycle[j])
                    if constraint_1 not in added_constraints and constraint_2 not in added_constraints:
                        prob += 0 <= y[(cycle[i], cycle[j])] + y[(cycle[j], cycle[k])] <= 1 + y[(cycle[i], cycle[k])]
                        added_constraints.add(constraint_1)
                        added_constraints.add(constraint_2)
    logging.info(f"Constraint hinzugefügt für: {(cycle)}")

    """
    # Entferne Duplikate aus dem Zyklus
    cycle = list(dict.fromkeys(cycle))
    counter = 0
    for i in range(len(cycle) - 1):
        if (cycle[i], cycle[i + 1]) in y:
            prob += y[(cycle[i], cycle[i + 1])] == 1
            counter += 1
        if (cycle[i + 1], cycle[i]) in y:
            prob += y[(cycle[i + 1], cycle[i])] == 0
            counter += 1
    # Letzte Kante zum Schließen des Zyklus behandeln
    if (cycle[-1], cycle[0]) in y:
        prob += y[(cycle[-1], cycle[0])] == 0
        counter += 1
    if (cycle[0], cycle[-1]) in y:
        prob += y[(cycle[0], cycle[-1])] == 1
        counter += 1
    logging.info(f"Es wurden {counter} Zyklusconstraints aufgestellt.")"""
"""
def add_cycle_constraints(prob, y, cycle):
    if cycle[-2] < cycle[-1] and (cycle[-2], cycle[-1]) in y and y[(cycle[-2], cycle[-1])] == 0:
        prob += y[(cycle[-2], cycle[-1])] == 1
    elif cycle[-1] < cycle[-2] and (cycle[-1], cycle[-2]) in y and y[(cycle[-1], cycle[-2])] == 0:
        prob += y[(cycle[-1], cycle[-2])] == 1
    elif cycle[-2] < cycle[-1] and (cycle[-2], cycle[-1]) in y and y[(cycle[-2], cycle[-1])] == 1:
        prob += y[(cycle[-2], cycle[-1])] == 0
    elif cycle[-1] < cycle[-2] and (cycle[-1], cycle[-2]) in y and y[(cycle[-1], cycle[-2])] == 1:
        prob += y[(cycle[-1], cycle[-2])] == 0

        
    logging.info(f"Cycle: {cycle}")
    for i in range(len(cycle) - 2):
        for j in range(i + 1, len(cycle)-1):
            if cycle[i] < cycle[j] and (cycle[i], cycle[j]) in y:
                prob += y[(cycle[i], cycle[j])] == 1
            elif cycle[j] < cycle[i] and (cycle[j], cycle[i]) in y:
                prob += y[(cycle[j], cycle[i])] == 1"""

def solve_bipartite_minimization(graph_file):
    start_time = time()
    logging.info(f"\n Prozess für {graph_file} gestartet")
    # Extrahiere den Basisnamen der Eingabedatei
    base_name = os.path.basename(graph_file)
    new_base_name = base_name.replace('.gr', '.sol')
    new_base_name1 = base_name.replace('.gr', '.cros')
    # Erstelle den Ausgabepfad
    output_file = os.path.join('mytests/solutions', new_base_name)
    output_file1 = os.path.join('mytests/crossings', new_base_name1)
    logging.info(f"Die Ausgabedatei wird {output_file} sein")

    edges = []
    with open(graph_file, "r") as file:
        for line in file:
            if line.startswith('c'):
                continue
            elif line.startswith('p'):
                parts = line.split()
                n0 = int(parts[2])  # Anzahl der Knoten in A
                n1 = int(parts[3])  # Anzahl der Knoten in B
                logging.info(f"Größen der Partitionen: A={n0}, B={n1}")
            else:
                x, y = map(int, line.split())
                edges.append((x, y))
    logging.info(f"{len(edges)} Kanten geladen.")

    prob = LpProblem("Minimize_Crossings", LpMinimize)

    # Boolesche Variablen für relative Positionen innerhalb jeder Partition
    y = {(i, j): LpVariable(f"y_{i}_{j}", 0, 1, cat='Binary') for i in range(n0 + 1, n0 + n1 + 1) for j in range(n0 + 1, n0 + n1 + 1) if i != j}
    c = {(i, j, k, l): LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary') for (i, j) in edges for (k, l) in edges}# if i < j and k < l and i < k and j != l}
    logging.info("x, y und c geladen.")
    
    # Zielfunktion, die minimiert werden soll
    prob += lpSum(c.values())
    logging.info("Zielfunktion aufgestellt.")
    """
    edges.sort(key=lambda x: x[0])  # Sortieren der Kanten nach dem Startknoten
    for idx, (i, j) in enumerate(edges):
        for (k, l) in edges[idx + 1:]:
    """
    for (i, j) in edges:
        for (k, l) in edges:
            if k > i:  # Da die Liste sortiert ist, brauchen wir nur k > i zu prüfen
                #if (i, j, k, l) not in c:
                    #c[(i, j, k, l)] = LpVariable(f"c_{i}_{j}_{k}_{l}", 0, 1, cat='Binary')
                if j > l:
                    prob += c[(i, j, k, l)] == y[(l, j)]
                elif l > j:
                    prob += c[(i, j, k, l)] == 1 - y[(j, l)]
    logging.info("Crossing Constraints aufgestellt.")
    #solved_cycles = set()
    i = 0
    added_constraints = set()
    while True:
        prob.solve()
        logging.info("Optimale Lösung gefunden. Ergebnisse werden gespeichert.")

        graph = defaultdict(list)
        in_degree = defaultdict(int)

        nodes = range(n0+1, n0+n1+1)
        for i in nodes:
            for j in nodes:
                if i != j:
                    y_ij = y.get((i, j))
                    
                    # Hinzufügen einer Kante basierend auf y_ij-Werten
                    if y_ij is not None:
                        if y_ij.value() == 1:
                            # i kommt vor j
                            graph[i].append(j)
                            in_degree[j] += 1
                        elif y_ij.value() == 0:
                            # j kommt vor i
                            graph[j].append(i)
                            in_degree[i] += 1
                            
        # Zyklusprüfung in den y-Variablen nach dem Lösen des ILP-Modells
        cycle_found, cycle_nodes = has_cycle(graph, nodes)
        if cycle_found:

            cycle_y_values = {(i, j): y[(i, j)].varValue for i in cycle_nodes for j in cycle_nodes if (i, j) in y}
            logging.error(f"Zyklus gefunden in den y-Variablen: {cycle_nodes} mit y-Werten: {cycle_y_values}")
        
            add_cycle_constraints(prob, y, cycle_nodes, added_constraints)
        else:
            break

    if prob.status == LpStatusOptimal:
        # Topologische Sortierung
        zero_in_degree_queue = deque([i for i in nodes if in_degree[i] == 0])
        sorted_b = []
        while zero_in_degree_queue:
            node = zero_in_degree_queue.popleft()
            sorted_b.append(node)
            for neighbor in graph[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)
        a = count_crossings_via_variables(c)
        # Ausgabe der sortierten Knoten
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w') as f:
            for b in sorted_b:
                f.write(f"{b}\n")
                print(f"{b}")
        print("Crossings: ", a) 
        logging.info(f"Ergebnisse in {output_file} gespeichert")
        # Ausgabe der sortierten Knoten

        a = count_crossings_via_variables(c)

        os.makedirs(os.path.dirname(output_file1), exist_ok=True)
        with open(output_file1, 'w') as f:
            f.write(f"Crossings: {a}\n")     
        logging.info(f"Crossings: {a}, in {output_file1} gespeichert")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  # 
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")
    else:
        logging.warning("Keine optimale Lösung gefunden.")

        end_time = time()
        elapsed_time = end_time - start_time
        elapsed_time_td = timedelta(seconds=elapsed_time)
        formatted_time = str(elapsed_time_td)  #
        formatted_time = formatted_time[:-3] 
        logging.info(f"Verstrichene Zeit: {formatted_time} (h:m:s.ms) \n")


def process_directory(directory_path):
    # Durchlaufe alle Dateien im angegebenen Verzeichnis
    for filename in os.listdir(directory_path):
        if filename.endswith('.gr'):  # Überprüfen, ob die Datei eine .gr Datei ist
            file_path = os.path.join(directory_path, filename)
            solve_bipartite_minimization(file_path)  # Rufe die Verarbeitungsfunktion für jede .gr Datei auf
            logging.info(f"Verarbeitung abgeschlossen für {file_path}")

#directory_path = 'githubtests/tiny_test_set/instances/'
#directory_path = 'mytests/instances/'
#process_directory(directory_path)

#test_file = 'githubtests/tiny_test_set/instances/complete_4_5.gr'
test_file = 'githubtests/medium_test_set/instances/10.gr'
#test_file = 'mytests/instances/1.gr'
solve_bipartite_minimization(test_file)
